from typing import Tuple

import orjson
import polars as pl

from bizon.common.models import SyncMetadata
from bizon.destination.destination import AbstractDestination
from bizon.engine.backend.backend import AbstractBackend
from bizon.monitoring.monitor import AbstractMonitor
from bizon.source.callback import AbstractSourceCallback

from .config import FileDestinationDetailsConfig


class FileDestination(AbstractDestination):
    def __init__(
        self,
        sync_metadata: SyncMetadata,
        config: FileDestinationDetailsConfig,
        backend: AbstractBackend,
        source_callback: AbstractSourceCallback,
        monitor: AbstractMonitor,
    ):
        super().__init__(sync_metadata, config, backend, source_callback, monitor)
        self.config: FileDestinationDetailsConfig = config

    def check_connection(self) -> bool:
        return True

    def delete_table(self) -> bool:
        return True

    def write_records(self, df_destination_records: pl.DataFrame) -> Tuple[bool, str]:
        if self.config.unnest:
            schema_keys = set([column.name for column in self.record_schemas[self.destination_id]])

            with open(f"{self.destination_id}.json", "a") as f:
                for value in [orjson.loads(data) for data in df_destination_records["source_data"].to_list()]:
                    assert set(value.keys()) == schema_keys, "Keys do not match the schema"

                    # Unnest the source_data column
                    row = {}
                    for column in self.record_schemas[self.destination_id]:
                        row[column.name] = value[column.name]

                    f.write(f"{orjson.dumps(row).decode('utf-8')}\n")

        else:
            df_destination_records.write_ndjson(f"{self.destination_id}.json")

        return True, ""



import pytest
from yaml import safe_load

from bizon.common.models import (
    BizonConfig,
    StreamConfig,
    StreamDestinationConfig,
    StreamSourceConfig,
)


class TestStreamSourceConfig:
    def test_kafka_topic_field(self):
        config = StreamSourceConfig(topic="events.users.update")
        assert config.topic == "events.users.update"

    def test_api_endpoint_field(self):
        config = StreamSourceConfig(endpoint="/api/v1/users")
        assert config.endpoint == "/api/v1/users"

    def test_extra_fields_allowed(self):
        config = StreamSourceConfig(topic="test", custom_field="custom_value")
        assert config.topic == "test"
        assert config.custom_field == "custom_value"


class TestStreamDestinationConfig:
    def test_table_id_required(self):
        config = StreamDestinationConfig(table_id="project.dataset.table")
        assert config.table_id == "project.dataset.table"

    def test_with_record_schema(self):
        config = StreamDestinationConfig(
            table_id="project.dataset.table",
            record_schema=[
                {"name": "id", "type": "STRING", "mode": "REQUIRED"},
                {"name": "data", "type": "JSON", "mode": "NULLABLE"},
            ],
            clustering_keys=["id"],
        )
        assert len(config.record_schema) == 2
        assert config.clustering_keys == ["id"]


class TestStreamConfig:
    def test_valid_stream_config(self):
        config = StreamConfig(
            name="user_events",
            source=StreamSourceConfig(topic="events.users.update"),
            destination=StreamDestinationConfig(table_id="project.dataset.users"),
        )
        assert config.name == "user_events"
        assert config.source.topic == "events.users.update"
        assert config.destination.table_id == "project.dataset.users"

    def test_table_id_format_validation_valid(self):
        config = StreamConfig(
            name="test",
            source=StreamSourceConfig(topic="test"),
            destination=StreamDestinationConfig(table_id="project.dataset.table"),
        )
        assert config.destination.table_id == "project.dataset.table"

    def test_table_id_format_validation_invalid(self):
        with pytest.raises(ValueError, match="table_id must be in format"):
            StreamConfig(
                name="test",
                source=StreamSourceConfig(topic="test"),
                destination=StreamDestinationConfig(table_id="invalid_table_id"),
            )

    def test_table_id_format_validation_two_parts(self):
        with pytest.raises(ValueError, match="Found 2 parts instead of 3"):
            StreamConfig(
                name="test",
                source=StreamSourceConfig(topic="test"),
                destination=StreamDestinationConfig(table_id="dataset.table"),
            )


class TestBizonConfigWithStreams:
    @pytest.fixture
    def base_config(self):
        return {
            "name": "test_pipeline",
            "source": {
                "name": "kafka",
                "stream": "topic",
                "sync_mode": "stream",
                "bootstrap_servers": "localhost:9092",
                "group_id": "test-group",
                "authentication": {
                    "type": "basic",
                    "params": {"username": "user", "password": "pass"},
                },
            },
            "destination": {
                "name": "logger",
                "config": {},
            },
            "engine": {"runner": {"type": "stream"}},
        }

    def test_streams_config_loads_successfully(self, base_config):
        base_config["streams"] = [
            {
                "name": "user_events",
                "source": {"topic": "events.users.update"},
                "destination": {"table_id": "project.dataset.users"},
            }
        ]
        config = BizonConfig(**base_config)
        assert len(config.streams) == 1
        assert config.streams[0].name == "user_events"

    def test_streams_requires_stream_sync_mode(self, base_config):
        base_config["source"]["sync_mode"] = "full_refresh"
        base_config["streams"] = [
            {
                "name": "test",
                "source": {"topic": "test"},
                "destination": {"table_id": "project.dataset.table"},
            }
        ]
        with pytest.raises(ValueError, match="requires source.sync_mode='stream'"):
            BizonConfig(**base_config)

    def test_duplicate_stream_names_rejected(self, base_config):
        base_config["streams"] = [
            {
                "name": "duplicate_name",
                "source": {"topic": "topic1"},
                "destination": {"table_id": "project.dataset.table1"},
            },
            {
                "name": "duplicate_name",
                "source": {"topic": "topic2"},
                "destination": {"table_id": "project.dataset.table2"},
            },
        ]
        with pytest.raises(ValueError, match="Duplicate stream names"):
            BizonConfig(**base_config)

    def test_duplicate_table_ids_rejected(self, base_config):
        base_config["streams"] = [
            {
                "name": "stream1",
                "source": {"topic": "topic1"},
                "destination": {"table_id": "project.dataset.same_table"},
            },
            {
                "name": "stream2",
                "source": {"topic": "topic2"},
                "destination": {"table_id": "project.dataset.same_table"},
            },
        ]
        with pytest.raises(ValueError, match="Duplicate table_ids"):
            BizonConfig(**base_config)

    def test_multiple_streams_valid(self, base_config):
        base_config["streams"] = [
            {
                "name": "stream1",
                "source": {"topic": "topic1"},
                "destination": {"table_id": "project.dataset.table1"},
            },
            {
                "name": "stream2",
                "source": {"topic": "topic2"},
                "destination": {"table_id": "project.dataset.table2"},
            },
        ]
        config = BizonConfig(**base_config)
        assert len(config.streams) == 2


class TestKafkaTopicsInjectionFromStreams:
    """Test that topics are injected into Kafka source config from streams.

    Note: BizonConfig.source is typed as SourceConfig (base class), so Kafka-specific
    fields like 'topics' aren't preserved after model instantiation. However, the
    injection still works at the dict level, and the Kafka connector factory
    properly instantiates KafkaSourceConfig with the injected topics.

    These tests verify the injection works by directly instantiating KafkaSourceConfig
    from the pre-processed data.
    """

    @pytest.fixture
    def kafka_config_without_topics(self):
        return {
            "name": "test_kafka_pipeline",
            "source": {
                "name": "kafka",
                "stream": "topic",
                "sync_mode": "stream",
                "bootstrap_servers": "localhost:9092",
                "group_id": "test-group",
                "authentication": {
                    "type": "basic",
                    "params": {"username": "user", "password": "pass"},
                },
            },
            "destination": {
                "name": "logger",
                "config": {},
            },
            "engine": {"runner": {"type": "stream"}},
            "streams": [
                {
                    "name": "user_events",
                    "source": {"topic": "events.users.update"},
                    "destination": {"table_id": "project.dataset.users"},
                },
                {
                    "name": "order_events",
                    "source": {"topic": "events.orders.create"},
                    "destination": {"table_id": "project.dataset.orders"},
                },
            ],
        }

    def test_topics_injected_from_streams(self, kafka_config_without_topics):
        from bizon.connectors.sources.kafka.src.config import KafkaSourceConfig

        # Run the model_validator by instantiating BizonConfig
        # This modifies the input dict in place
        BizonConfig(**kafka_config_without_topics)

        # Verify topics were injected into source dict
        source_dict = kafka_config_without_topics["source"]
        topics = source_dict.get("topics")
        assert topics is not None
        assert len(topics) == 2

        topic_names = [t["name"] for t in topics]
        assert "events.users.update" in topic_names
        assert "events.orders.create" in topic_names

        # Verify KafkaSourceConfig can be instantiated with injected topics
        kafka_config = KafkaSourceConfig(**source_dict)
        assert len(kafka_config.topics) == 2

    def test_topics_not_overwritten_if_already_set(self):
        from bizon.connectors.sources.kafka.src.config import KafkaSourceConfig

        config_dict = {
            "name": "test_kafka_pipeline",
            "source": {
                "name": "kafka",
                "stream": "topic",
                "sync_mode": "stream",
                "bootstrap_servers": "localhost:9092",
                "group_id": "test-group",
                "topics": [{"name": "existing.topic", "destination_id": "existing.dest"}],
                "authentication": {
                    "type": "basic",
                    "params": {"username": "user", "password": "pass"},
                },
            },
            "destination": {
                "name": "logger",
                "config": {},
            },
            "engine": {"runner": {"type": "stream"}},
            "streams": [
                {
                    "name": "new_stream",
                    "source": {"topic": "new.topic"},
                    "destination": {"table_id": "project.dataset.new"},
                },
            ],
        }

        # Run the model_validator
        BizonConfig(**config_dict)

        # Verify existing topics were not overwritten
        source_dict = config_dict["source"]
        topics = source_dict.get("topics")
        assert len(topics) == 1
        assert topics[0]["name"] == "existing.topic"

        # Verify KafkaSourceConfig preserves the existing topic
        kafka_config = KafkaSourceConfig(**source_dict)
        assert len(kafka_config.topics) == 1
        assert kafka_config.topics[0].name == "existing.topic"


class TestRecordSchemasInjectionFromStreams:
    @pytest.fixture
    def bigquery_config_with_streams(self):
        return {
            "name": "test_bq_pipeline",
            "source": {
                "name": "kafka",
                "stream": "topic",
                "sync_mode": "stream",
                "bootstrap_servers": "localhost:9092",
                "group_id": "test-group",
                "authentication": {
                    "type": "basic",
                    "params": {"username": "user", "password": "pass"},
                },
            },
            "destination": {
                "name": "bigquery_streaming_v2",
                "config": {
                    "project_id": "test-project",
                    "dataset_id": "test_dataset",
                    "unnest": True,
                },
            },
            "engine": {"runner": {"type": "stream"}},
            "streams": [
                {
                    "name": "user_events",
                    "source": {"topic": "events.users.update"},
                    "destination": {
                        "table_id": "project.dataset.users",
                        "record_schema": [
                            {"name": "id", "type": "STRING", "mode": "REQUIRED"},
                            {"name": "data", "type": "JSON", "mode": "NULLABLE"},
                        ],
                        "clustering_keys": ["id"],
                    },
                },
            ],
        }

    def test_record_schemas_injected_from_streams(self, bigquery_config_with_streams):
        config = BizonConfig(**bigquery_config_with_streams)

        # Check that record_schemas were injected
        record_schemas = config.destination.config.record_schemas
        assert record_schemas is not None
        assert len(record_schemas) == 1
        assert record_schemas[0].destination_id == "project.dataset.users"
        assert len(record_schemas[0].record_schema) == 2
        assert record_schemas[0].clustering_keys == ["id"]

    def test_unnest_true_works_with_streams_record_schemas(self, bigquery_config_with_streams):
        # This should not raise validation error because record_schemas are injected
        config = BizonConfig(**bigquery_config_with_streams)
        assert config.destination.config.unnest is True
        assert config.destination.config.record_schemas is not None


class TestStreamsConfigFromYaml:
    def test_full_yaml_config(self):
        from bizon.connectors.sources.kafka.src.config import KafkaSourceConfig

        yaml_config = """
        name: kafka_to_bigquery_with_streams

        source:
          name: kafka
          stream: topic
          sync_mode: stream
          bootstrap_servers: localhost:9092
          group_id: test-group
          authentication:
            type: basic
            params:
              username: user
              password: pass

        destination:
          name: bigquery_streaming_v2
          config:
            project_id: my-project
            dataset_id: my_dataset
            unnest: true

        streams:
          - name: user_events
            source:
              topic: events.users.update
            destination:
              table_id: my-project.my_dataset.users
              record_schema:
                - name: id
                  type: STRING
                  mode: REQUIRED
                - name: email
                  type: STRING
                  mode: NULLABLE
              clustering_keys:
                - id

          - name: order_events
            source:
              topic: events.orders.create
            destination:
              table_id: my-project.my_dataset.orders
              record_schema:
                - name: order_id
                  type: INTEGER
                  mode: REQUIRED
                - name: total
                  type: FLOAT
                  mode: NULLABLE
              clustering_keys:
                - order_id

        engine:
          runner:
            type: stream
        """
        config_dict = safe_load(yaml_config)
        config = BizonConfig(**config_dict)

        # Verify streams loaded correctly
        assert len(config.streams) == 2
        assert config.streams[0].name == "user_events"
        assert config.streams[0].source.topic == "events.users.update"
        assert config.streams[1].name == "order_events"

        # Verify topics were injected into source dict
        # (BizonConfig.source is SourceConfig base class, so we check the raw dict)
        source_dict = config_dict["source"]
        topics = source_dict.get("topics")
        assert len(topics) == 2

        # Verify KafkaSourceConfig can parse the injected topics
        kafka_config = KafkaSourceConfig(**source_dict)
        assert len(kafka_config.topics) == 2
        topic_names = [t.name for t in kafka_config.topics]
        assert "events.users.update" in topic_names
        assert "events.orders.create" in topic_names

        # Verify record_schemas were injected
        assert len(config.destination.config.record_schemas) == 2



import random
import time
from multiprocessing import Queue
from typing import Union

from loguru import logger

from bizon.destination.destination import AbstractDestination
from bizon.engine.queue.config import QUEUE_TERMINATION, QueueMessage
from bizon.engine.queue.queue import AbstractQueue, AbstractQueueConsumer
from bizon.monitoring.monitor import AbstractMonitor
from bizon.source.models import SourceIteration
from bizon.transform.transform import Transform

from .config import PythonQueueConfigDetails
from .consumer import PythonQueueConsumer


class PythonQueue(AbstractQueue):
    def __init__(self, config: PythonQueueConfigDetails, **kwargs) -> None:
        super().__init__(config)
        self.config: PythonQueueConfigDetails = config

        assert "queue" in kwargs, "queue param passed in kwargs is required for PythonQueue"
        self.queue: Queue = kwargs["queue"]

    def connect(self):
        # No connection to establish for PythonQueue
        pass

    def get_consumer(
        self,
        destination: AbstractDestination,
        transform: Transform,
        monitor: AbstractMonitor,
    ) -> AbstractQueueConsumer:
        return PythonQueueConsumer(
            config=self.config,
            queue=self.queue,
            destination=destination,
            transform=transform,
            monitor=monitor,
        )

    def put_queue_message(self, queue_message: QueueMessage):
        if not self.queue.full():
            self.queue.put(queue_message)
            logger.debug(f"Putting data from iteration {queue_message.iteration} items in queue)")
        else:
            logger.warning("Queue is full, waiting for consumer to consume data")
            time.sleep(random.random())
            self.put_queue_message(queue_message)

    def get(self) -> QueueMessage:
        if not self.queue.empty():
            queue_message: QueueMessage = self.queue.get()
            logger.debug(f"Got {len(queue_message.df_source_records.height)} records from queue")
            return queue_message
        else:
            logger.debug("Queue is empty, waiting for producer to produce data")
            time.sleep(random.random())
            return self.get()

    def get_size(self) -> Union[int, None]:
        if hasattr(self.queue, "qsize"):
            return self.queue.qsize()
        return None

    def terminate(self, iteration: int) -> bool:
        self.put(
            source_iteration=SourceIteration(next_pagination={}, records=[]),
            iteration=iteration,
            signal=QUEUE_TERMINATION,
        )
        logger.info("Sent termination signal to destination.")
        return True



from datetime import datetime
from typing import List

from loguru import logger
from polars import DataFrame
from pytz import UTC

from .models import destination_record_schema


class DestinationBuffer:
    def __init__(self, buffer_size: int, buffer_flush_timeout: int) -> None:
        self.buffer_size = buffer_size * 1024 * 1024  # Convert to bytes
        self.buffer_flush_timeout = buffer_flush_timeout
        self.df_destination_records: DataFrame = DataFrame(schema=destination_record_schema)
        self._iterations: List[int] = []
        self.pagination = {}
        self.modified_at: List[datetime] = [datetime.now(tz=UTC)]

    @property
    def is_empty(self) -> bool:
        """Check if buffer is empty"""
        return self.df_destination_records.height == 0

    @property
    def current_size(self) -> int:
        """Return buffer size"""
        return self.df_destination_records.estimated_size(unit="b")

    @property
    def buffer_free_space_pct(self) -> float:
        """Return free space in buffer in percentage"""
        return round((self.buffer_free_space / self.buffer_size) * 100, 3)

    @property
    def from_iteration(self) -> int:
        """Return the smallest iteration in buffer"""
        if not self._iterations:
            raise ValueError("Buffer is empty")
        return min(self._iterations)

    @property
    def to_iteration(self) -> int:
        """Return the largest iteration in buffer"""
        if not self._iterations:
            raise ValueError("Buffer is empty")
        return max(self._iterations)

    @property
    def buffer_free_space(self) -> int:
        """Return free space for records in buffer"""
        assert self.current_size <= self.buffer_size, "Buffer size exceeded"
        return self.buffer_size - self.current_size

    @property
    def ripeness(self) -> float:
        """Return buffer ripeness"""
        if self.buffer_flush_timeout == 0:
            return 0
        return round((max(self.modified_at) - min(self.modified_at)).seconds, 2)

    @property
    def is_ripe(self) -> bool:
        """Check if buffer is ripe for flushing based on the timeout"""
        if self.buffer_flush_timeout == 0:
            return False
        return (max(self.modified_at) - min(self.modified_at)).seconds >= self.buffer_flush_timeout

    def flush(self):
        """Flush buffer"""
        self.df_destination_records = DataFrame(schema=destination_record_schema)
        self._iterations = []
        self.pagination = {}
        self.modified_at = []

    def add_source_iteration_records_to_buffer(
        self, iteration: int, df_destination_records: DataFrame, pagination: dict = None
    ):
        """Add records for the given iteration to buffer"""
        self.df_destination_records.vstack(df_destination_records, in_place=True)
        self._iterations.append(iteration)
        self.pagination = pagination
        self.modified_at.append(datetime.now(tz=UTC))

        logger.info(
            f"Added {df_destination_records.height} records to buffer for iteration {iteration} - {self.df_destination_records.estimated_size(unit='mb')} MB"
        )



from enum import Enum
from typing import Union

from pydantic import BaseModel, Field

from .authenticators.basic import BasicHttpAuthParams
from .authenticators.cookies import CookiesAuthParams
from .authenticators.oauth import Oauth2AuthParams
from .authenticators.token import TokenAuthParams


class AuthType(str, Enum):
    OAUTH = "oauth"
    API_KEY = "api_key"
    BEARER = "bearer"
    BASIC = "basic"
    COOKIES = "cookies"


class AuthConfig(BaseModel):
    type: AuthType = Field(..., description="Type of authentication", example="oauth")
    params: Union[
        Oauth2AuthParams,
        BasicHttpAuthParams,
        TokenAuthParams,
        CookiesAuthParams,
    ] = Field(..., description="Configuration for the authentication")



from typing import List, Tuple, Type

from google.cloud.bigquery import SchemaField
from google.cloud.bigquery_storage_v1.types import ProtoSchema
from google.protobuf.descriptor_pb2 import (
    DescriptorProto,
    FieldDescriptorProto,
    FileDescriptorProto,
)
from google.protobuf.descriptor_pool import DescriptorPool
from google.protobuf.message import Message
from google.protobuf.message_factory import GetMessageClassesForFiles


def map_bq_type_to_field_descriptor(bq_type: str) -> int:
    """Map BigQuery type to Protobuf FieldDescriptorProto type."""
    type_map = {
        "STRING": FieldDescriptorProto.TYPE_STRING,  # STRING -> TYPE_STRING
        "BYTES": FieldDescriptorProto.TYPE_BYTES,  # BYTES -> TYPE_BYTES
        "INTEGER": FieldDescriptorProto.TYPE_INT64,  # INTEGER -> TYPE_INT64
        "FLOAT": FieldDescriptorProto.TYPE_DOUBLE,  # FLOAT -> TYPE_DOUBLE
        "NUMERIC": FieldDescriptorProto.TYPE_STRING,  # NUMERIC -> TYPE_STRING (use string to handle precision)
        "BIGNUMERIC": FieldDescriptorProto.TYPE_STRING,  # BIGNUMERIC -> TYPE_STRING
        "BOOLEAN": FieldDescriptorProto.TYPE_BOOL,  # BOOLEAN -> TYPE_BOOL
        "DATE": FieldDescriptorProto.TYPE_STRING,  # DATE -> TYPE_STRING
        "DATETIME": FieldDescriptorProto.TYPE_STRING,  # DATETIME -> TYPE_STRING
        "TIME": FieldDescriptorProto.TYPE_STRING,  # TIME -> TYPE_STRING
        "TIMESTAMP": FieldDescriptorProto.TYPE_STRING,  # TIMESTAMP -> TYPE_INT64 (Unix epoch time)
        "RECORD": FieldDescriptorProto.TYPE_MESSAGE,  # RECORD -> TYPE_MESSAGE (nested message)
    }

    return type_map.get(bq_type, FieldDescriptorProto.TYPE_STRING)  # Default to TYPE_STRING


def get_proto_schema_and_class(bq_schema: List[SchemaField]) -> Tuple[ProtoSchema, Type[Message]]:
    """Generate a ProtoSchema and a TableRow class for unnested BigQuery schema."""
    # Define the FileDescriptorProto
    file_descriptor_proto = FileDescriptorProto()
    file_descriptor_proto.name = "dynamic.proto"
    file_descriptor_proto.package = "dynamic_package"

    # Define the TableRow message schema
    message_descriptor = DescriptorProto()
    message_descriptor.name = "TableRow"

    # Add fields to the message, only use TYPE_STRING, BigQuery does not support other types
    # It does not imapact data types in final table

    # https://stackoverflow.com/questions/70489919/protobuf-type-for-bigquery-timestamp-field
    fields = [
        {
            "name": col.name,
            "type": map_bq_type_to_field_descriptor(col.field_type),
            "label": (
                FieldDescriptorProto.LABEL_REQUIRED if col.mode == "REQUIRED" else FieldDescriptorProto.LABEL_OPTIONAL
            ),
        }
        for col in bq_schema
    ]

    for i, field in enumerate(fields, start=1):
        field_descriptor = message_descriptor.field.add()
        field_descriptor.name = field["name"]
        field_descriptor.number = i
        field_descriptor.type = field["type"]
        field_descriptor.label = field["label"]

    # Add the message to the file descriptor
    file_descriptor_proto.message_type.add().CopyFrom(message_descriptor)

    # Create a DescriptorPool and register the FileDescriptorProto
    pool = DescriptorPool()
    pool.Add(file_descriptor_proto)

    # Use the registered file name to fetch the message classes
    message_classes = GetMessageClassesForFiles(["dynamic.proto"], pool=pool)

    # Fetch the TableRow class
    table_row_class = message_classes["dynamic_package.TableRow"]

    # Create the ProtoSchema
    proto_schema = ProtoSchema()
    proto_schema.proto_descriptor.CopyFrom(message_descriptor)

    return proto_schema, table_row_class



from enum import Enum
from typing import Dict, Optional

from pydantic import BaseModel, Field


class MonitorType(str, Enum):
    DATADOG = "datadog"


class BaseMonitoringConfig(BaseModel):
    enable_tracing: bool = Field(default=False, description="Enable tracing for the monitor")


class DatadogConfig(BaseMonitoringConfig):
    datadog_agent_host: Optional[str] = None
    datadog_host_env_var: Optional[str] = None
    datadog_agent_port: int = 8125
    tags: Optional[Dict[str, str]] = Field(default={}, description="Key-value pairs to add to the monitor as tags")

    @property
    def host_is_configured(self) -> bool:
        return bool(self.datadog_agent_host or self.datadog_host_env_var)

    def __init__(self, **data):
        super().__init__(**data)
        if not self.host_is_configured:
            raise ValueError("Either datadog_agent_host or datadog_host_env_var must be specified")

    class Config:
        extra = "forbid"


class MonitoringConfig(BaseMonitoringConfig):
    type: MonitorType
    config: Optional[DatadogConfig] = None

    class Config:
        extra = "forbid"



import io
import struct
from functools import cache
from typing import Tuple, Union

import fastavro
from avro.schema import Schema
from confluent_kafka.serialization import SerializationError

# Constants for schema ID byte sizes
APICURIO_SCHEMA_ID_BYTES = 8
CONFLUENT_SCHEMA_ID_BYTES = 4
MAGIC_BYTE = 0


class Hashabledict(dict):
    """A hashable dictionary for caching purposes"""

    def __hash__(self):
        return hash(frozenset(self.items()))


@cache
def parse_global_id_from_serialized_message(message: bytes) -> Tuple[int, int]:
    """
    Parse the global id from the serialized message.

    Args:
        message: The serialized message bytes

    Returns:
        Tuple of (schema_id, number_of_bytes_used_for_schema_id)

    Raises:
        SerializationError: If the message is invalid or missing schema id
    """
    size = len(message)

    if size < CONFLUENT_SCHEMA_ID_BYTES + 1:
        raise SerializationError("Invalid message. Missing schema id")

    # Create BytesIO object for easier reading
    message_buffer = io.BytesIO(message)
    message_buffer.seek(0)
    magic_byte = message_buffer.read(1)

    if magic_byte != bytes([MAGIC_BYTE]):
        raise SerializationError(
            f"Unexpected magic byte {magic_byte}. This message was not produced with a Schema Registry serializer"
        )

    # Read Confluent schema ID (4 bytes + 1 magic byte)
    message_buffer.seek(0)
    schema_id = struct.unpack(">bI", message_buffer.read(CONFLUENT_SCHEMA_ID_BYTES + 1))[1]

    # If schema_id is 0, try reading as Apicurio format (8 bytes)
    if schema_id == 0:
        if size < APICURIO_SCHEMA_ID_BYTES + 1:
            raise SerializationError("Invalid Apicurio message. Missing schema id")
        message_buffer.seek(0)
        schema_id = struct.unpack(">bq", message_buffer.read(APICURIO_SCHEMA_ID_BYTES + 1))[1]
        return schema_id, APICURIO_SCHEMA_ID_BYTES
    else:
        return schema_id, CONFLUENT_SCHEMA_ID_BYTES


def decode_avro_message(message_value: bytes, nb_bytes_schema_id: int, avro_schema: Union[Schema, dict]) -> dict:
    """
    Decode an Avro message.

    Args:
        message_value: The raw message bytes
        nb_bytes_schema_id: Number of bytes used for schema ID
        avro_schema: The Avro schema (as Schema object or dict)

    Returns:
        Decoded message as a dictionary
    """
    # Create BytesIO from message bytes
    message_bytes = io.BytesIO(message_value)

    # Skip magic byte and schema ID bytes
    message_bytes.seek(nb_bytes_schema_id + 1)

    # Decode the message using fastavro
    if isinstance(avro_schema, Schema):
        schema_dict = avro_schema.to_json()
    else:
        schema_dict = avro_schema

    data = fastavro.schemaless_reader(message_bytes, schema_dict)

    return data



import concurrent.futures
import time
from threading import Event

from loguru import logger

from bizon.common.models import BizonConfig
from bizon.engine.pipeline.models import PipelineReturnStatus
from bizon.engine.runner.config import RunnerStatus
from bizon.engine.runner.runner import AbstractRunner


class ThreadRunner(AbstractRunner):
    def __init__(self, config: BizonConfig):
        super().__init__(config)

    # TODO: refacto this
    def get_kwargs(self):
        extra_kwargs = {}

        if self.bizon_config.engine.queue.type == "python_queue":
            from queue import Queue

            queue = Queue(maxsize=self.bizon_config.engine.queue.config.queue.max_size)
            extra_kwargs["queue"] = queue

        return extra_kwargs

    def run(self) -> RunnerStatus:
        """Run the pipeline with dedicated threads for source and destination"""

        extra_kwargs = self.get_kwargs()
        job = AbstractRunner.init_job(bizon_config=self.bizon_config, config=self.config, **extra_kwargs)

        # Store the future results
        result_producer = None
        result_consumer = None

        # Start the producer and consumer events
        producer_stop_event = Event()
        consumer_stop_event = Event()

        extra_kwargs = self.get_kwargs()

        with concurrent.futures.ThreadPoolExecutor(
            max_workers=self.bizon_config.engine.runner.config.max_workers
        ) as executor:
            future_producer = executor.submit(
                AbstractRunner.instanciate_and_run_producer,
                self.bizon_config,
                self.config,
                job.id,
                producer_stop_event,
                **extra_kwargs,
            )
            logger.info("Producer thread has started ...")

            time.sleep(self.bizon_config.engine.runner.config.consumer_start_delay)

            future_consumer = executor.submit(
                AbstractRunner.instanciate_and_run_consumer,
                self.bizon_config,
                self.config,
                job.id,
                consumer_stop_event,
                **extra_kwargs,
            )
            logger.info("Consumer thread has started ...")

            while future_producer.running() and future_consumer.running():
                logger.debug("Producer and consumer are still running ...")
                self._is_running = True
                time.sleep(self.bizon_config.engine.runner.config.is_alive_check_interval)

            self._is_running = False

            if not future_producer.running():
                result_producer: PipelineReturnStatus = future_producer.result()
                logger.info(f"Producer thread stopped running with result: {result_producer}")

                if result_producer.SUCCESS:
                    logger.info("Producer thread has finished successfully, will wait for consumer to finish ...")
                else:
                    logger.error("Producer thread failed, stopping consumer ...")
                    consumer_stop_event.set()

            if not future_consumer.running():
                result_consumer = future_consumer.result()
                logger.info(f"Consumer thread stopped running with result: {result_consumer}")

                if result_consumer == PipelineReturnStatus.SUCCESS:
                    logger.info("Consumer thread has finished successfully")
                else:
                    logger.error("Consumer thread failed, stopping producer ...")
                    producer_stop_event.set()

        runner_status = RunnerStatus(producer=future_producer.result(), consumer=future_consumer.result())

        if not runner_status.is_success:
            logger.error(runner_status.to_string())

        return runner_status



from abc import ABC
from enum import Enum
from typing import Optional

from pydantic import BaseModel, ConfigDict, Field, field_validator


class DestinationTypes(str, Enum):
    BIGQUERY = "bigquery"
    BIGQUERY_STREAMING = "bigquery_streaming"
    BIGQUERY_STREAMING_V2 = "bigquery_streaming_v2"
    LOGGER = "logger"
    FILE = "file"


class DestinationColumn(BaseModel, ABC):
    name: str = Field(..., description="Name of the column")
    type: str = Field(..., description="Type of the column")
    description: Optional[str] = Field(None, description="Description of the column")


class RecordSchemaConfig(BaseModel):
    # Forbid extra keys in the model
    model_config = ConfigDict(extra="forbid")

    destination_id: str = Field(..., description="Destination ID")
    record_schema: list[DestinationColumn] = Field(..., description="Record schema")


class AbstractDestinationDetailsConfig(BaseModel):
    # Forbid extra keys in the model
    model_config = ConfigDict(extra="forbid")

    buffer_size: int = Field(
        default=50,
        description="Buffer size in Mb for the destination. Set to 0 to disable and write directly to the destination.",
    )

    buffer_flush_timeout: int = Field(
        default=600,
        description="Maximum time in seconds for buffering after which the records will be written to the destination. Set to 0 to deactivate the timeout buffer check.",  # noqa
    )

    max_concurrent_threads: int = Field(
        default=10,
        description="Maximum number of concurrent threads to use for writing to the destination.",
    )

    record_schemas: Optional[list[RecordSchemaConfig]] = Field(
        default=None, description="Schemas for the records. Required if unnest is set to true."
    )

    unnest: bool = Field(
        default=False,
        description="Unnest the data before writing to the destination. Schema should be provided in the model_config.",
    )

    authentication: Optional[BaseModel] = Field(
        description="Authentication configuration for the destination, if needed", default=None
    )

    destination_id: Optional[str] = Field(
        description="Destination ID, identifier to use to store the records in the destination", default=None
    )

    @field_validator("unnest", mode="before")
    def validate_record_schema_if_unnest(cls, value, values):
        if bool(value) and not values.data.get("record_schemas", []):
            raise ValueError("At least one `record_schemas` must be provided if `unnest` is set to True.")
        return value


class AbstractDestinationConfig(BaseModel):
    # Forbid extra keys in the model
    model_config = ConfigDict(extra="forbid")

    name: DestinationTypes = Field(..., description="Name of the destination")
    alias: str = Field(
        ...,
        description="Alias of the destination, used for tracking the system name (ie bigquery for bigquery_streaming)",
    )
    config: AbstractDestinationDetailsConfig = Field(..., description="Configuration for the destination")



from typing import Tuple

import polars as pl
from loguru import logger

from bizon.common.models import SyncMetadata
from bizon.destination.destination import AbstractDestination
from bizon.engine.backend.backend import AbstractBackend
from bizon.monitoring.monitor import AbstractMonitor
from bizon.source.callback import AbstractSourceCallback

from .config import LoggerDestinationConfig


class LoggerDestination(AbstractDestination):
    def __init__(
        self,
        sync_metadata: SyncMetadata,
        config: LoggerDestinationConfig,
        backend: AbstractBackend,
        source_callback: AbstractSourceCallback,
        monitor: AbstractMonitor,
    ):
        super().__init__(
            sync_metadata=sync_metadata,
            config=config,
            backend=backend,
            source_callback=source_callback,
            monitor=monitor,
        )

    def check_connection(self) -> bool:
        return True

    def delete_table(self) -> bool:
        return True

    def write_records(self, df_destination_records: pl.DataFrame) -> Tuple[bool, str]:
        for record in df_destination_records.iter_rows(named=True):
            logger.info(record["source_data"])
        return True, ""



from datetime import datetime

import polars as pl
from pytz import UTC

from bizon.destination.models import transform_to_df_destination_records
from bizon.source.models import source_record_schema

df_source_records = pl.DataFrame(
    {
        "id": ["record_1", "record_2"],
        "data": ['{"key": "value1"}', '{"key": "value2"}'],
        "timestamp": [datetime(2024, 12, 5, 11, 30, tzinfo=UTC), datetime(2024, 12, 5, 12, 30, tzinfo=UTC)],
        "destination_id": ["test", "test"],
    },
    schema=source_record_schema,
)


def test_destination_record_from_source_record():
    destination_source_records = transform_to_df_destination_records(
        df_source_records=df_source_records,
        extracted_at=datetime(2024, 12, 5, 12, 0),
    )
    assert destination_source_records["source_record_id"].to_list() == ["record_1", "record_2"]
    assert destination_source_records["source_data"].to_list() == ['{"key": "value1"}', '{"key": "value2"}']



from enum import Enum
from typing import Literal, Optional

import polars as pl
from pydantic import BaseModel, Field

from bizon.destination.config import (
    AbstractDestinationConfig,
    AbstractDestinationDetailsConfig,
    DestinationColumn,
    DestinationTypes,
)


class GCSBufferFormat(str, Enum):
    PARQUET = "parquet"
    CSV = "csv"


class TimePartitioning(str, Enum):
    DAY = "DAY"
    HOUR = "HOUR"
    MONTH = "MONTH"
    YEAR = "YEAR"


class BigQueryColumnType(str, Enum):
    BOOLEAN = "BOOLEAN"
    BYTES = "BYTES"
    DATE = "DATE"
    DATETIME = "DATETIME"
    FLOAT = "FLOAT"
    FLOAT64 = "FLOAT64"
    GEOGRAPHY = "GEOGRAPHY"
    INTEGER = "INTEGER"
    INT64 = "INT64"
    NUMERIC = "NUMERIC"
    BIGNUMERIC = "BIGNUMERIC"
    JSON = "JSON"
    RECORD = "RECORD"
    STRING = "STRING"
    TIME = "TIME"
    TIMESTAMP = "TIMESTAMP"


class BigQueryColumnMode(str, Enum):
    NULLABLE = "NULLABLE"
    REQUIRED = "REQUIRED"
    REPEATED = "REPEATED"


BIGQUERY_TO_POLARS_TYPE_MAPPING = {
    "STRING": pl.String,
    "BYTES": pl.Binary,
    "INTEGER": pl.Int64,
    "INT64": pl.Int64,
    "FLOAT": pl.Float64,
    "FLOAT64": pl.Float64,
    "NUMERIC": pl.Float64,  # Can be refined for precision with Decimal128 if needed
    "BIGNUMERIC": pl.Float64,  # Similar to NUMERIC
    "BOOLEAN": pl.Boolean,
    "BOOL": pl.Boolean,
    "TIMESTAMP": pl.String,  # We use BigQuery internal parsing to convert to datetime
    "DATE": pl.String,  # We use BigQuery internal parsing to convert to datetime
    "DATETIME": pl.String,  # We use BigQuery internal parsing to convert to datetime
    "TIME": pl.Time,
    "GEOGRAPHY": pl.Object,  # Polars doesn't natively support geography types
    "ARRAY": pl.List,  # Requires additional handling for element types
    "JSON": pl.String,
}


class BigQueryColumn(DestinationColumn):
    name: str = Field(..., description="Name of the column")
    type: BigQueryColumnType = Field(..., description="Type of the column")
    mode: BigQueryColumnMode = Field(..., description="Mode of the column")
    description: Optional[str] = Field(None, description="Description of the column")
    default_value_expression: Optional[str] = Field(None, description="Default value expression")

    @property
    def polars_type(self):
        return BIGQUERY_TO_POLARS_TYPE_MAPPING.get(self.type.upper())


class BigQueryAuthentication(BaseModel):
    service_account_key: str = Field(
        description="Service Account Key JSON string. If empty it will be infered",
        default="",
    )


class BigQueryRecordSchemaConfig(BaseModel):
    destination_id: str = Field(..., description="Destination ID")
    record_schema: list[BigQueryColumn] = Field(..., description="Record schema")

    # BigQuery Clustering Keys
    clustering_keys: Optional[list[str]] = Field(None, description="Clustering keys")


class BigQueryConfigDetails(AbstractDestinationDetailsConfig):
    # Table details
    project_id: str = Field(..., description="BigQuery Project ID")
    dataset_id: str = Field(..., description="BigQuery Dataset ID")
    dataset_location: str = Field(default="US", description="BigQuery Dataset location")

    # GCS Buffer
    gcs_buffer_bucket: str = Field(..., description="GCS Buffer bucket")
    gcs_buffer_format: GCSBufferFormat = Field(default=GCSBufferFormat.PARQUET, description="GCS Buffer format")

    # Time partitioning
    time_partitioning: TimePartitioning = Field(
        default=TimePartitioning.DAY, description="BigQuery Time partitioning type"
    )

    # Schema for unnesting
    record_schemas: Optional[list[BigQueryRecordSchemaConfig]] = Field(
        default=None, description="Schema for the records. Required if unnest is set to true."
    )

    authentication: Optional[BigQueryAuthentication] = None


class BigQueryConfig(AbstractDestinationConfig):
    name: Literal[DestinationTypes.BIGQUERY]
    alias: str = "bigquery"
    buffer_size: Optional[int] = 400
    config: BigQueryConfigDetails



import time
from datetime import datetime

import polars as pl

from bizon.destination.buffer import DestinationBuffer
from bizon.destination.models import destination_record_schema

TEST_BUFFER_SIZE = 1  # Mb
TEST_BUFFER_TIMEOUT = 1


def test_buffer_destination():
    buffer = DestinationBuffer(buffer_size=TEST_BUFFER_SIZE, buffer_flush_timeout=60)
    assert buffer.buffer_size == TEST_BUFFER_SIZE * 1024 * 1024
    assert buffer._iterations == []


df_destination_records = pl.DataFrame(
    {
        "bizon_id": ["id_1", "id_2"],
        "bizon_extracted_at": [datetime(2024, 12, 5, 12, 0), datetime(2024, 12, 5, 13, 0)],
        "bizon_loaded_at": [datetime(2024, 12, 5, 12, 30), datetime(2024, 12, 5, 13, 30)],
        "source_record_id": ["record_1", "record_2"],
        "source_timestamp": [datetime(2024, 12, 5, 11, 30), datetime(2024, 12, 5, 12, 30)],
        "source_data": ["cookies", "cream"],
    },
    schema=destination_record_schema,
)


def test_buffer_records():
    buffer = DestinationBuffer(buffer_size=TEST_BUFFER_SIZE, buffer_flush_timeout=60)
    buffer.add_source_iteration_records_to_buffer(iteration=0, df_destination_records=df_destination_records)
    assert buffer.df_destination_records.equals(df_destination_records)
    assert buffer._iterations == [0]
    assert buffer.from_iteration == 0 and buffer.to_iteration == 0


def test_buffer_flush():
    buffer = DestinationBuffer(buffer_size=TEST_BUFFER_SIZE, buffer_flush_timeout=60)

    buffer.add_source_iteration_records_to_buffer(iteration=0, df_destination_records=df_destination_records)

    assert buffer.df_destination_records.equals(df_destination_records)

    buffer.flush()

    assert buffer.df_destination_records.height == 0


def test_buffer_iterations():
    buffer = DestinationBuffer(buffer_size=TEST_BUFFER_SIZE, buffer_flush_timeout=TEST_BUFFER_TIMEOUT)
    buffer.add_source_iteration_records_to_buffer(iteration=0, df_destination_records=df_destination_records)
    buffer.add_source_iteration_records_to_buffer(iteration=1, df_destination_records=df_destination_records)

    assert len(buffer._iterations) == 2

    buffer.add_source_iteration_records_to_buffer(iteration=2, df_destination_records=df_destination_records)

    assert len(buffer._iterations) == 3
    assert buffer.from_iteration == 0 and buffer.to_iteration == 2
    buffer.flush()


def test_buffer_ripeness():
    buffer = DestinationBuffer(buffer_size=2000, buffer_flush_timeout=TEST_BUFFER_TIMEOUT)
    buffer.add_source_iteration_records_to_buffer(iteration=0, df_destination_records=df_destination_records)
    time.sleep(1)
    buffer.add_source_iteration_records_to_buffer(iteration=1, df_destination_records=df_destination_records)
    assert buffer.ripeness <= TEST_BUFFER_TIMEOUT
    assert buffer.is_ripe



from contextlib import contextmanager
from typing import Dict

from bizon.common.models import SyncMetadata
from bizon.engine.pipeline.models import PipelineReturnStatus
from bizon.monitoring.config import MonitoringConfig
from bizon.monitoring.monitor import AbstractMonitor


class NoOpMonitor(AbstractMonitor):
    def __init__(self, sync_metadata: SyncMetadata, monitoring_config: MonitoringConfig):
        super().__init__(sync_metadata, monitoring_config)

    def track_pipeline_status(self, pipeline_status: PipelineReturnStatus) -> None:
        pass

    @contextmanager
    def trace(self, operation_name: str, resource: str = None, extra_tags: Dict[str, str] = None):
        """
        No-op trace implementation.

        Args:
            operation_name (str): The name of the operation being traced
            resource (str): The resource being operated on (e.g., topic name, table name)
            extra_tags (Dict[str, str]): Additional tags for the trace

        Yields:
            None (no-op implementation)
        """
        yield None
